#include "max-in-sub-array.h"

#include<iostream>
#include<iomanip>
#include<cstdlib>
#include<ctime>

using namespace std;

template<typename T>
void make_random_array(T* ptr, size_t size) {
  for (size_t i = 0; i < size; i++)
    ptr[i] = static_cast<T>(rand());
}

template<typename T>
T make_random_integer(T a, T b)
{
  T result(a + rand() % (b - a + 1));
  return result;
}

double time()
{
  timespec t;
  clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &t);
  return t.tv_sec + 1e-9 * t.tv_nsec;
}

template<typename T, int options>
double benchmark_one_path(size_t buffer_size, size_t update_size, size_t draw_size, size_t draws_between_updates)
{
  buffer<T, options> b;

  T *array = new T[buffer_size];
  make_random_array(array, buffer_size);
  b.bufferData(array, buffer_size);

  T dummy = 0;

  double minimum_run_time = 0.1;
  size_t repeat = 1;
  double run_time = 0.0;

  while (true) {

    double start_time = time();

    for (size_t i = 0; i < repeat; i++) {
      size_t update_offset = make_random_integer<size_t>(0, buffer_size - update_size);

      make_random_array(array, update_size);
      b.bufferSubData(update_offset, array, update_size);

      for (size_t j = 0; j < draws_between_updates; j++) {
        size_t draw_offset = make_random_integer<size_t>(0, buffer_size - draw_size);
        dummy += b.max(draw_offset, draw_size);
      }
    }

    double end_time = time();

    run_time = end_time - start_time;

    if (run_time >= minimum_run_time ||
        repeat >= 100000000)
    {
      break;
    }
    else
    {
      repeat *= 4;
    }
  }


  delete[] array;

  if (dummy == 123 && run_time == 123.456) abort(); // just to guarantee it's not optimized out

  return run_time / repeat;
}

void benchmark(size_t buffer_size, size_t update_size, size_t draw_size, size_t draws_between_updates, double* time, double* speedup)
{
  double smart = benchmark_one_path<unsigned short, SMART_MAX>(buffer_size, update_size, draw_size, draws_between_updates);
  double naive = benchmark_one_path<unsigned short, NAIVE_MAX>(buffer_size, update_size, draw_size, draws_between_updates);
  *time = smart;
  *speedup = naive/smart;
}

int main()
{
  cout.precision(3);
  cout << setw(14) << "buffer update" << setw(14) << "number of" << setw(14) << "draw call" << setw(14) << "time" << setw(14) << "speedup" << endl;
  cout << setw(14) << "size" << setw(14) << "draw calls" << setw(14) << "size" << setw(14) << "(s)" << setw(14) << "(x)" << endl;

  size_t buffer_size = 1048576;
  for (size_t draws_between_updates = 1;
          draws_between_updates <= 65536;
          draws_between_updates *= 16)
  {
    for (size_t draw_size = 4; draw_size <= std::min(buffer_size, size_t(65536)); draw_size *= 4)
    {
      for (size_t update_size = 4;
           update_size <= std::min(buffer_size, draw_size * draws_between_updates);
           update_size *= 4)
      {
        double time, speedup;
        benchmark(buffer_size,
                  std::min(update_size, size_t(0.9 * buffer_size)),
                  std::min(draw_size, size_t(0.9 * buffer_size)),
                  draws_between_updates,
                  &time,
                  &speedup
                 );
        cout << setw(14) << update_size
             << setw(14) << draws_between_updates
             << setw(14) << draw_size
             << setw(14) << time
             << setw(14) << speedup
             << endl;
      }
    }
  }
}
